/**
 *
 * fileName : expr_interpreter.h
 */
#include "SablotConfig.h"	//@@[Sablot configurations]

#include "expr.h"
#include "verts.h"
#include <float.h>
#include <math.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/wait.h>

#include "context.h"
#include "tree.h"
#include "utf8.h"
#include "domprovider.h"
#include "guard.h"

#ifdef _MSC_VER
#	define	FILE_READ_MODE	"rt"
#else
#	define	FILE_READ_MODE	"r"
#endif
#define EXFF_INTERPRETER_MAX_TREE_DEPTH		32

#if !SABLOTCONFIG_DELETE_XPATH_INTERPRETER//@@[Delete XPath interpreter]

/**
 * s1s2񂪂邩Bstrstr()Ɠ.
 */
static char* strstr_ex(const char* s1, const char* s2, int size)
{
	const char* p1 = s1;
	const char* p2 = s2;
	int i = 0;
	int strl = strlen(s2);

	while((*p1 && *p2) && i <= (size - strl))
	{
		if (*p1 == *p2)
		{
			p1++;
			p2++;
		}
		else
		{
			p1 -= p2 - s2 -1;
			p2 = s2;
			i++;
		}
	}

	return (*p2 ? NULL: (char*)(p1 - (p2 - s2)));
}


/**
 * &lt; &gt; &amp; &apos; &quot; ꂼ < > & ' " ɓWJăRs[
 * dest̃obt@TCY͂Ȃ炸\ɂƂĂ
 * 	nStr:	src̕ ( <= dest̃obt@TCY)
 */
static char* strncpy_ex(char* dest, const char* src, size_t nStr)
{
	if (nStr <= 0) return dest;

	// &amp; &apos; &gt; &lt; &quot; & ' > < " ͑SASKIIȂ̂UTF-8łR[h
	switch (*src) {
	case '&':
		switch(*(src+1)) {
		case 'a':
			if (*(src+2)=='m' && *(src+3)=='p' && *(src+4)==';') {	// &amp;
				*dest = '&';
				*(dest+1) = '\0';
				strncpy_ex(dest+1, src+5, nStr-5);
				return dest;
			} else if (*(src+2)=='p' && *(src+3)=='o' && *(src+4)=='s' && *(src+5)==';') {	// &apos;
				*dest = '\'';
				*(dest+1) = '\0';
				strncpy_ex(dest+1, src+6, nStr-6);
				return dest;
			}
			break;
		case 'g':
			if (*(src+2)=='t' && *(src+3)==';') {	// &gt;
				*dest = '>';
				*(dest+1) = '\0';
				strncpy_ex(dest+1, src+4, nStr-4);
				return dest;
			}
			break;
		case 'l':
			if (*(src+2)=='t' && *(src+3)==';') {	// &lt;
				*dest = '<';
				*(dest+1) = '\0';
				strncpy_ex(dest+1, src+4, nStr-4);
				return dest;
			}
			break;
		case 'q':
			if (*(src+2)=='u' && *(src+3)=='o' && *(src+4)=='t' && *(src+5)==';') {	// &quot;
				*dest = '\"';
				*(dest+1) = '\0';
				strncpy_ex(dest+1, src+6, nStr-6);
				return dest;
			}
			break;
		default:
			break;
		}
		break;
	default:
		break;
	}

	// 1Rs[
	*dest = *src;
	*(dest+1) = '\0';
	strncpy_ex(dest+1, src+1, nStr-1);
	return dest;

}

/**
 * |{
 * @param S
 * @param retxpr
 * @param atoms
 * @param c
 * @param retPtn
 *       KEYTRNS_RETPTN_KEY : |󏈗ʂȂꍇ͌IDԂ
 *       KEYTRNS_RETPTN_NULL : |󏈗ʂȂꍇ͋󕶎ԂB
 * @return eFlag OK .
 *               NOT_OK s.
 */
eFlag Expression::key_translate(Sit S, Expression &retxpr, ExprList &atoms, Context *c, int retPtn)
{
	NodeHandle newroot = NULL;
	assert(S.domExternal() || proc);
	GP( Context ) newc = new Context;

	const FileMmap *fileMmap;

	int atomsNumber = atoms.number();
	if (atomsNumber != 2)
	{
		Err(S, ET_BAD_ARGS_N);
	}

	Str resName, resKey;
	E( atoms[0] -> tostring(S, resName) );
	E( atoms[1] -> tostring(S, resKey ) );

	Processor* proc = S.getProcessor();
    Tree *newtree;

    Str absolute = resName, scheme; // Path.
	makeAbsoluteURI2(resName, proc->baseForVertex(&getOwnerElement()), absolute, scheme);

	fileMmap = S.getFileMmap(absolute);

	if(!fileMmap) {
		Err1(S, E_FILE_OPEN, absolute);
		return NOT_OK;
	}
	
	// 'pID'̐. pID݂̂ƑIDqbg\邽.
	char* resKeyName = (char*)malloc(strlen(resKey) + 10);
	if (!resKeyName)
	{
		return NOT_OK;
	}
	sprintf(resKeyName, "key=\'%s\'", (const char*)resKey);

	// pID݂邩ۂ.
	char* tmp = strstr_ex(fileMmap->m_map, resKeyName, fileMmap->size);
	if (tmp == NULL)
	{
		free (resKeyName);
		if( retPtn == KEYTRNS_RETPTN_NULL ){
			retxpr.setAtom("");
		}
		else{
			retxpr.setAtom(resKey);
		}

		return OK;
	}

	// pID̕񕪃|C^ړ.
	tmp = tmp + strlen(resKeyName);

	// ŏ[']text̍ŏ̈ʒuɂȂ.
	char* buf = strchr(tmp, '\'');
	if (buf == NULL)
	{
		free (resKeyName);
		return NOT_OK;
	}

	char* pStart = buf+1;	// text̊Jnʒu.
	char* pEnd = strchr(pStart, '\''); // text̏Iʒu.
	char *text= (char*)malloc(pEnd-pStart+1);
	// textobt@ɃRs[.
	strncpy_ex(text, pStart, pEnd - pStart);

	retxpr.setAtom(text);
	free (resKeyName);
	free (text);

    return OK;
}

/**
 * XPATH֐interpreteȑ
 *   |󏈗ŌʂȂꍇ͌pIDԂI/F
 * @param S
 * @param retxpr
 * @param atoms
 * @param c
 * @return eFlag OK .
 *               NOT_OK s.
 */
eFlag Expression::callFunc_interpreter(Sit S, Expression &retxpr, ExprList &atoms, Context *c)
{
	return key_translate(S, retxpr, atoms, c, KEYTRNS_RETPTN_KEY);
}

/**
 * XPATH֐interpreter_compatiblȅ
 *   |󏈗ŌʂȂꍇ͋󕶎ԂI/F
 * @param S
 * @param retxpr
 * @param atoms
 * @param c
 * @return eFlag OK .
 *               NOT_OK s.
 */
eFlag Expression::callFunc_interpreter_compatible(Sit S, Expression &retxpr, ExprList &atoms, Context *c)
{
	return key_translate(S, retxpr, atoms, c, KEYTRNS_RETPTN_NULL);
}

#endif
